package service

import (
	"context"
	"errors"
	"fmt"
	json "github.com/json-iterator/go"
	"github.com/olivere/elastic/v7"
	"github.com/opensearch-project/opensearch-go/opensearchapi"
	"github.com/spf13/cast"
	"github.com/thoas/go-funk"
	"gitlab.wodcloud.com/smart-operation/so-operation-api/src/bean/entity"
	"gitlab.wodcloud.com/smart-operation/so-operation-api/src/bean/vo/request"
	"gitlab.wodcloud.com/smart-operation/so-operation-api/src/bean/vo/response"
	"gitlab.wodcloud.com/smart-operation/so-operation-api/src/common/client"
	"gitlab.wodcloud.com/smart-operation/so-operation-api/src/common/conf"
	"gitlab.wodcloud.com/smart-operation/so-operation-api/src/pkg/beagle/jsontime"
	"go.uber.org/zap"
	"io"
	"log"
	"net/http"
	"strings"
	"time"
)

type AlertSvc struct {
	User entity.SystemUserInfo
}

var (
	OpenSearchIndex = "so_alert"
	Mapping         = strings.NewReader(`{
    "settings": {
        "number_of_shards": 1,
        "number_of_replicas": 0,
        "index.max_result_window": "1000000"
    },
    "mappings": {
        "properties": {
            "id": {
                "type": "integer"
            },
            "alert_point": {
                "type": "text",
                "fields": {
                    "keyword": {
                        "type": "keyword",
                        "ignore_above": 256
                    }
                }
            },
            "alert_rules_id": {
                "type": "keyword"
            },
            "risk_level": {
                "type": "integer"
            },
            "alert_time": {
                "type": "date",
                "format": "yyyy-MM-dd HH:mm:ss||yyyy-MM-dd||epoch_millis",
                "ignore_malformed": true
            },
            "class_id": {
                "type": "integer"
            },
            "class_parent_name": {
                "type": "text",
                "fields": {
                    "keyword": {
                        "type": "keyword",
                        "ignore_above": 256
                    }
                }
            },
            "class_name": {
                "type": "text",
                "fields": {
                    "keyword": {
                        "type": "keyword",
                        "ignore_above": 256
                    }
                }
            },
            "metric_config_id": {
                "type": "keyword"
            },
            "metric_config_name": {
                "type": "text",
                "fields": {
                    "keyword": {
                        "type": "keyword",
                        "ignore_above": 256
                    }
                }
            },
            "alert_rule_type": {
                "type": "keyword"
            },
            "alert_rule_type_name": {
                "type": "text",
                "fields": {
                    "keyword": {
                        "type": "keyword",
                        "ignore_above": 256
                    }
                }
            },
            "current_value": {
                "type": "integer"
            },
            "notification_count": {
                "type": "integer"
            },
            "push_count": {
                "type": "integer"
            },
            "last_push_time": {
                "type": "date",
                "format": "yyyy-MM-dd HH:mm:ss||yyyy-MM-dd||epoch_millis",
                "ignore_malformed": true
            },
            "status": {
                "type": "integer"
            },
            "push_records": {
                "type": "nested",
                "properties": {
                    "id": {
                        "type": "integer"
                    },
                    "alert_id": {
                        "type": "integer"
                    },
                    "alert_rules_id": {
                        "type": "keyword"
                    },
                    "risk_level": {
                        "type": "integer"
                    },
                    "notify_method": {
                        "type": "keyword"
                    },
                    "system_account": {
                        "type": "keyword"
                    },
                    "push_type": {
                        "type": "integer"
                    },
                    "status": {
                        "type": "integer"
                    },
                    "user_name": {
                        "type": "text",
                        "fields": {
                            "keyword": {
                                "type": "keyword",
                                "ignore_above": 256
                            }
                        }
                    },
                    "phone": {
                        "type": "keyword"
                    },
                    "created_by": {
                        "type": "keyword"
                    },
                    "updated_by": {
                        "type": "keyword"
                    }
                }
            },
            "disposed_list": {
                "type": "nested",
                "properties": {
                    "disposal_content": {
                        "type": "text",
                        "fields": {
                            "keyword": {
                                "type": "keyword",
                                "ignore_above": 256
                            }
                        }
                    },
                    "is_disposed": {
                        "type": "integer"
                    },
                    "disposal_user": {
                        "type": "keyword"
                    },
                    "disposal_time": {
                        "type": "date",
                        "format": "yyyy-MM-dd HH:mm:ss||yyyy-MM-dd||epoch_millis",
                        "ignore_malformed": true
                    }
                }
            },
            "is_disposed": {
                "type": "integer"
            },
            "close_remark": {
                "type": "text",
                "fields": {
                    "keyword": {
                        "type": "keyword",
                        "ignore_above": 256
                    }
                }
            },
            "close_user": {
                "type": "text",
                "fields": {
                    "keyword": {
                        "type": "keyword",
                        "ignore_above": 256
                    }
                }
            },
            "close_time": {
                "type": "text",
                "fields": {
                    "keyword": {
                        "type": "keyword",
                        "ignore_above": 256
                    }
                }
            },
            "defer_push": {
                "type": "integer"
            },
            "alert_condition": {
                "properties": {
                    "thresholds_min": {
                        "type": "integer"
                    },
                    "thresholds_max": {
                        "type": "integer"
                    },
                    "risk_level": {
                        "type": "integer"
                    }
                }
            },
            "disposal_content": {
                "type": "text",
                "fields": {
                    "keyword": {
                        "type": "keyword",
                        "ignore_above": 256
                    }
                }
            },
            "created_by": {
                "type": "keyword"
            },
            "created_at": {
                "type": "date",
                "format": "yyyy-MM-dd HH:mm:ss||yyyy-MM-dd||epoch_millis",
                "ignore_malformed": true
            },
            "updated_by": {
                "type": "keyword"
            },
            "updated_at": {
                "type": "date",
                "format": "yyyy-MM-dd HH:mm:ss||yyyy-MM-dd||epoch_millis",
                "ignore_malformed": true
            }
        }
    }
}`)
)

func (a *AlertSvc) CreateIndex() error {
	cli, err := client.GetOpenSearch()
	if err != nil {
		return err
	}
	res := opensearchapi.IndicesCreateRequest{
		Index: OpenSearchIndex,
		Body:  Mapping,
	}
	do, err := res.Do(context.Background(), cli)
	if err != nil {
		return err
	}
	defer do.Body.Close()
	if do.StatusCode != http.StatusOK {
		return errors.New(do.String())
	}
	return nil
}

func (a *AlertSvc) DeleteIndex() error {
	cli, err := client.GetOpenSearch()
	if err != nil {
		return err
	}
	res := opensearchapi.IndicesDeleteRequest{
		Index: []string{OpenSearchIndex},
	}
	do, err := res.Do(context.Background(), cli)
	if err != nil {
		return err
	}
	defer do.Body.Close()
	if do.StatusCode != http.StatusOK {
		return errors.New(do.String())
	}
	return nil
}

func (a *AlertSvc) Indices() (indices []string, err error) {
	var (
		catIndicesList []response.CatIndices
	)

	cli, err := client.GetOpenSearch()
	if err != nil {
		return
	}
	res := opensearchapi.CatIndicesRequest{Format: "json"}
	do, err := res.Do(context.Background(), cli)
	if err != nil {
		return
	}
	defer do.Body.Close()
	if do.StatusCode != http.StatusOK {
		err = errors.New(do.String())
		return
	}
	body, err := io.ReadAll(do.Body)
	if err != nil {
		return
	}

	err = json.Unmarshal(body, &catIndicesList)
	if err != nil {
		return
	}
	funkGet := funk.Get(catIndicesList, "Index") //取一层
	indices, ok := funkGet.([]string)
	if !ok {
		err = errors.New("funk.Get failed")
		return
	}
	return
}

func (a *AlertSvc) IndexSearch(req request.ListAlert) (resp response.AlertList, err error) {
	var (
		sources response.OpenSearchSource
	)

	cli, err := client.GetOpenSearch()
	if err != nil {
		return
	}

	boolQuery := elastic.NewBoolQuery()

	if req.Id != 0 {
		boolQuery.Must(elastic.NewMatchQuery("id", req.Id))
	}

	if req.RiskLevel != 0 {
		boolQuery.Must(elastic.NewMatchQuery("risk_level", req.RiskLevel))
	}

	if req.Status != 0 {
		boolQuery.Must(elastic.NewMatchQuery("status", req.Status))
	}

	// 请输入预警点/分类/指标
	if req.Keyword != "" {
		subBoolQuery := elastic.NewBoolQuery()
		subBoolQuery.Should(elastic.NewMultiMatchQuery(req.Keyword, "alert_point", "class_parent_name", "class_name", "metric_config_name"))
		//subBoolQuery.Should(elastic.NewMatchQuery("class_name", req.Keyword))
		boolQuery.Must(subBoolQuery)
	}

	if req.StartTime != "" {
		boolQuery.Filter(elastic.NewRangeQuery("created_at").Gte(req.StartTime).Lte(req.EndTime))
	}

	querySource, _ := boolQuery.Source()
	b, _ := json.Marshal(querySource)
	log.Printf("es statements: %s\n", string(b))

	if len(req.Ids) > 0 {
		var idsStr []string
		for _, id := range req.Ids {
			idsStr = append(idsStr, cast.ToString(id))
		}
		ids := entity.OpenSearchIds{
			Ids: struct {
				Values []string `json:"values"`
			}{idsStr},
		}
		b, err = json.Marshal(&ids)
		if err != nil {
			return
		}
	}

	content := strings.NewReader(fmt.Sprintf(`{
  "query": %s,
  "from": %d,
  "size": %d}`, string(b), req.GetPageSize()*(req.GetPage()-1), req.GetPageSize()))

	res := opensearchapi.SearchRequest{
		Index: []string{OpenSearchIndex},
		Body:  content,
		Sort:  []string{"id"},
	}
	do, err := res.Do(context.Background(), cli)
	if err != nil {
		return
	}
	defer do.Body.Close()
	if do.StatusCode != http.StatusOK {
		err = errors.New(do.String())
		return
	}
	body, err := io.ReadAll(do.Body)
	if err != nil {
		return
	}

	err = json.Unmarshal(body, &sources)
	if err != nil {
		return
	}

	for _, hit := range sources.Hits.Hits {
		resp.List = append(resp.List, hit.Source)
	}
	return
}

func (a *AlertSvc) IndexDocExist(req request.ExistAlert) (exist bool, err error) {
	var (
		sources response.OpenSearchSource
	)

	cli, err := client.GetOpenSearch()
	if err != nil {
		return
	}

	boolQuery := elastic.NewBoolQuery()
	if req.Id != 0 {
		boolQuery.Must(elastic.NewTermQuery("id", req.Id))
	}
	if req.AlertRulesId != "" {
		boolQuery.Must(elastic.NewTermQuery("alert_rules_id", req.AlertRulesId))
	}

	querySource, _ := boolQuery.Source()
	b, _ := json.Marshal(querySource)

	content := strings.NewReader(fmt.Sprintf(`{
  "query": %s,
  "from": %d,
  "size": %d}`, string(b), 0, 1))

	res := opensearchapi.SearchRequest{
		Index: []string{OpenSearchIndex},
		Body:  content,
		Sort:  []string{"id"},
	}
	do, err := res.Do(context.Background(), cli)
	if err != nil {
		return
	}
	defer do.Body.Close()
	if do.StatusCode != http.StatusOK {
		err = errors.New(do.String())
		return
	}
	body, err := io.ReadAll(do.Body)
	if err != nil {
		return
	}

	err = json.Unmarshal(body, &sources)
	if err != nil {
		return
	}

	if sources.Hits.Total.Value == 1 {
		exist = true
	}
	return
}

func (a *AlertSvc) IndexUpdate(req request.UpdateAlert) (err error) {
	var (
		sources response.OpenSearchSource
		now     = jsontime.Now()
	)

	cli, err := client.GetOpenSearch()
	if err != nil {
		return
	}

	doc := make(map[string]interface{}, 0)
	if req.CloseRemark != "" {
		doc["close_remark"] = req.CloseRemark
	}
	if req.DeferPush != 0 {
		doc["defer_push"] = req.DeferPush
	}
	if req.RiskLevel != 0 {
		doc["risk_level"] = req.RiskLevel
	}
	if req.Status != 0 {
		doc["status"] = req.Status
	}
	doc["updated_at"] = now

	if req.DisposalContent != "" {
		var detailAlert response.AlertItem
		detailAlert, err = a.GetDataById(request.DetailAlert{Id: req.Id})
		if err != nil {
			return err
		}
		disposedList := detailAlert.DisposedList
		disposedList = append(disposedList, entity.DisposedList{
			IsDisposed:      req.Status,
			DisposalContent: req.DisposalContent,
			DisposalUser:    a.User.SystemAccount,
			DisposalTime:    now,
		})
		doc["disposed_list"] = disposedList
	}

	if len(req.PushRecords) > 0 {
		doc["push_records"] = req.PushRecords
	}

	if len(doc) == 0 {
		return errors.New("no updated fields")
	}

	b, _ := json.Marshal(doc)
	docStr := string(b)
	if docStr == "" {
		return errors.New("doc marshal failed")
	}

	content := strings.NewReader(fmt.Sprintf(`{
  "doc": %s
}`, docStr))

	res := opensearchapi.UpdateRequest{
		Index:      OpenSearchIndex,
		DocumentID: cast.ToString(req.Id),
		Body:       content,
		Source:     []string{"true"},
	}
	do, err := res.Do(context.Background(), cli)
	if err != nil {
		return
	}
	defer do.Body.Close()
	if do.StatusCode != http.StatusOK {
		err = errors.New(do.String())
		return
	}
	body, err := io.ReadAll(do.Body)
	if err != nil {
		return
	}

	err = json.Unmarshal(body, &sources)
	if err != nil {
		return
	}

	conf.Logger.Info("--->alert update--->", zap.Any("IndexUpdate", sources))
	return
}

func (a *AlertSvc) Update(req request.UpdateAlert) error {
	err := a.IndexUpdate(req)
	return err
}

func (a *AlertSvc) BatchPushAlert(req request.BatchPushAlert) (err error) {
	now := jsontime.Now()
	resp, err := a.IndexSearch(request.ListAlert{Ids: req.Ids})
	if err != nil {
		return
	}

	for _, alert := range resp.List {
		pushRecords := alert.PushRecords
		pushCount := len(alert.PushRecords)
		for _, v := range req.NotifyRecipients {
			pushCount++
			data := entity.PushRecord{
				Id:            pushCount,
				AlertId:       alert.Id,
				AlertRulesId:  alert.AlertRulesId,
				RiskLevel:     alert.RiskLevel,
				NotifyMethod:  req.NotifyMethod,
				SystemAccount: v.SystemAccount,
				PushTime:      now,
				PushType:      2,
				Status:        0, // TODO 需要调用之后更新
				UserName:      v.UserName,
				Phone:         v.Phone,
				CreatedBy:     a.User.SystemAccount,
				CreatedAt:     now,
				UpdatedBy:     a.User.SystemAccount,
				UpdatedAt:     now,
			}
			pushRecords = append(pushRecords, data)
		}

		err = a.IndexUpdate(request.UpdateAlert{
			Id:          alert.Id,
			PushRecords: pushRecords,
		})
		if err != nil {
			return
		}
	}

	// TODO 批量推送用户告警
	conf.Logger.Info("batch push", zap.Any("payload", req))
	time.Sleep(time.Second)
	return nil
}

func (a *AlertSvc) BatchCloseAlert(req request.BatchCloseAlert) (err error) {
	var ids []int
	if len(req.Ids) > 0 {
		ids = req.Ids
	} else {
		ids = append(ids, req.Id)
	}
	for _, id := range ids {
		err = a.IndexUpdate(request.UpdateAlert{
			Id:          id,
			CloseRemark: req.CloseRemark,
			DeferPush:   req.DeferPush,
			Status:      3,
		})
		if err != nil {
			return
		}
	}
	conf.Logger.Info("batch close", zap.Any("payload", req))
	time.Sleep(time.Second)
	return
}

func (a *AlertSvc) GetDataById(req request.DetailAlert) (resp response.AlertItem, err error) {
	list, err := a.IndexSearch(request.ListAlert{Id: req.Id})
	if len(list.List) > 0 {
		resp = list.List[0]
	}
	return
}

func (a *AlertSvc) List(req request.ListAlert) (resp response.AlertList, err error) {
	resp, err = a.IndexSearch(req)
	resp.TotalCount = int64(len(resp.List))
	return
}

func (a *AlertSvc) DisposeAlert(req request.DisposeAlert) (err error) {
	// TODO 我的预警工单处置
	var (
		sources     response.OpenSearchSource
		now         = jsontime.Now()
		detailAlert response.AlertItem
	)

	detailAlert, err = a.GetDataById(request.DetailAlert{Id: req.Id})
	if err != nil {
		return err
	}

	if detailAlert.Status == 1 {
		return errors.New("alert has been restored")
	}
	if detailAlert.Status == 3 {
		return errors.New("alert has been closed")
	}

	cli, err := client.GetOpenSearch()
	if err != nil {
		return
	}

	doc := make(map[string]interface{}, 0)
	doc["status"] = req.Status
	if req.DisposalContent != "" {
		disposedList := detailAlert.DisposedList
		disposedList = append(disposedList, entity.DisposedList{
			IsDisposed:      req.Status,
			DisposalContent: req.DisposalContent,
			DisposalUser:    a.User.SystemAccount,
			DisposalTime:    now,
		})
		doc["disposed_list"] = disposedList
	}

	b, _ := json.Marshal(doc)
	docStr := string(b)
	if docStr == "" {
		return errors.New("doc marshal failed")
	}

	content := strings.NewReader(fmt.Sprintf(`{
  "doc": %s
}`, docStr))

	res := opensearchapi.UpdateRequest{
		Index:      OpenSearchIndex,
		DocumentID: cast.ToString(req.Id),
		Body:       content,
		Source:     []string{"true"},
	}
	do, err := res.Do(context.Background(), cli)
	if err != nil {
		return
	}
	defer do.Body.Close()
	if do.StatusCode != http.StatusOK {
		err = errors.New(do.String())
		return
	}
	body, err := io.ReadAll(do.Body)
	if err != nil {
		return
	}

	err = json.Unmarshal(body, &sources)
	if err != nil {
		return
	}
	time.Sleep(time.Second)
	return nil
}
